
#line 1 "drawing.vvh"
#ifndef DRAWING_VVH
#define DRAWING_VVH

#include <GL/glew.h>
#include "vve.h"
#include <util/gl.h>
#include <util/glu.h>
#include <vector>
#include <util/color.h>
#include <QMutex>
#include <QMutexLocker>
#include <QTextStream>
#include <shape/quadric.h>

#include <stdio.h>

#ifdef __APPLE__
#  define _GLUfuncptr void (*)()
#endif

static QTextStream out(stdout);
static QTextStream err(stderr);

template <typename Tissue, typename Model>
struct TissueDrawer
{
  IMPORT_COMPLEX_TYPES(Tissue);

  enum VBO_IDS
  {
    VBO_POSITIONS,
    VBO_NORMALS,
    VBO_CELL_COLORS,
    VBO_CONTOUR_COLORS,
    VBO_CELL_INDICES,
    VBO_CONTOUR_INDICES,
    NUMBER_VBOS
  };

  typedef util::Colorf Colorf;

  TissueDrawer(Tissue& t, Model *m)
    : model(m)
    , T(&t)
    , recompute(false)
    , initialized(false)
  { }

  void setTissue(Tissue& t)
  {
    if(&t != T)
    {
      T = &t;
      createVA();
    }
  }

  void updateGeometry()
  {
    createVA();
  }

  void updateColors()
  {
    recolorVA();
  }

  void createVA()
  {
    pmin = Point3d(HUGE_VAL, HUGE_VAL, HUGE_VAL);
    pmax = -pmin;
    QMutexLocker locker(&mutex);
    cell_indices.clear();
    contour_indices.clear();
    combined.clear();

    for(typeof(util::ForAll::range(T->W)) rangeb054404f_b4b4_11e1_a33a_bcaec51648ac = util::ForAll::range(T->W) ; !rangeb054404f_b4b4_11e1_a33a_bcaec51648ac.atEnd() ; ++rangeb054404f_b4b4_11e1_a33a_bcaec51648ac) {
            const junction& j = *rangeb054404f_b4b4_11e1_a33a_bcaec51648ac.current();
#line 78 "drawing.vvh"
{
      if(pmin.x() > j->pos.x()) pmin.x() = j->pos.x();
      if(pmin.y() > j->pos.y()) pmin.y() = j->pos.y();
      if(pmin.z() > j->pos.z()) pmin.z() = j->pos.z();
      if(pmax.x() < j->pos.x()) pmax.x() = j->pos.x();
      if(pmax.y() < j->pos.y()) pmax.y() = j->pos.y();
      if(pmax.z() < j->pos.z()) pmax.z() = j->pos.z();
    }}

    // First check the number of cell/point
    GLuint nb_junctions = 0;
    for(typeof(util::ForAll::range(T->C)) rangeb0559fe1_b4b4_11e1_83fb_bcaec51648ac = util::ForAll::range(T->C) ; !rangeb0559fe1_b4b4_11e1_83fb_bcaec51648ac.atEnd() ; ++rangeb0559fe1_b4b4_11e1_83fb_bcaec51648ac) {
            const cell& c = *rangeb0559fe1_b4b4_11e1_83fb_bcaec51648ac.current();
#line 90 "drawing.vvh"
nb_junctions += T->S.valence(c);}

    int nb_vtx = (T->drawBorders ? 2*nb_junctions : nb_junctions);

    cell_index.resize(nb_vtx);
    positions.resize(nb_vtx);
    normals.resize(nb_vtx);
    cell_indices.reserve(nb_vtx);
    cell_colors.resize(nb_vtx);
    if(T->drawBorders)
    {
      contour_indices.reserve(nb_vtx);
      contour_colors.resize(nb_vtx);
    }
    GLUtesselator *tess_contour = createTesselator(false);
    GLUtesselator *tess_cell = createTesselator(true);

    GLuint cur_pos = 0;
    for(typeof(util::ForAll::range(T->C)) rangeb055c6f0_b4b4_11e1_a0ea_bcaec51648ac = util::ForAll::range(T->C) ; !rangeb055c6f0_b4b4_11e1_a0ea_bcaec51648ac.atEnd() ; ++rangeb055c6f0_b4b4_11e1_a0ea_bcaec51648ac) {
            const cell& c = *rangeb055c6f0_b4b4_11e1_a0ea_bcaec51648ac.current();
#line 109 "drawing.vvh"
{
      if(T->drawBorders)
        drawCellWithContour(tess_cell, tess_contour, c, cur_pos);
      else
        drawCellNoContour(tess_cell, c, cur_pos);
      combined.clear();
    }}

    gluDeleteTess(tess_cell);
    gluDeleteTess(tess_contour);

    // Now, invert the array
    cell_positions.clear();
    for(GLuint i = 0 ; i < cell_index.size() ; ++i)
    {
      cell_positions.insert(std::make_pair(cell_index[i], i));
    }
    nb_cells = cell_indices.size();
    nb_contours = contour_indices.size();
    recompute = true;
  }

  void recolorVA()
  {
    QMutexLocker locker(&mutex);
    for(typeof(util::ForAll::range(T->C)) rangeb055ee00_b4b4_11e1_a15c_bcaec51648ac = util::ForAll::range(T->C) ; !rangeb055ee00_b4b4_11e1_a15c_bcaec51648ac.atEnd() ; ++rangeb055ee00_b4b4_11e1_a15c_bcaec51648ac) {
            const cell& c = *rangeb055ee00_b4b4_11e1_a15c_bcaec51648ac.current();
#line 135 "drawing.vvh"
{
      Colorf col = model->cellColor(c);
      Colorf con = model->contourColor(c);
      typename std::unordered_multimap<cell,int>::iterator first, last;
      for(typeof(util::ForAll::range(cell_positions.equal_range(c))) rangeb055ee01_b4b4_11e1_8eb5_bcaec51648ac = util::ForAll::range(cell_positions.equal_range(c)) ; !rangeb055ee01_b4b4_11e1_8eb5_bcaec51648ac.atEnd() ; ++rangeb055ee01_b4b4_11e1_8eb5_bcaec51648ac) {
            std::pair<cell,GLuint> pi = *rangeb055ee01_b4b4_11e1_8eb5_bcaec51648ac.current();
#line 140 "drawing.vvh"
{
        GLuint i = pi.second;
        cell_colors[i] = col;
        if(T->drawBorders)
          contour_colors[i] = con;
      }}
    }}
    recolor = true;
  }

  GLUtesselator *createTesselator(bool forCell)
  {
    GLUtesselator *tess = gluNewTess();
    gluTessCallback(tess, GLU_TESS_ERROR, (_GLUfuncptr)&tessError);
    if(forCell)
    {
      gluTessCallback(tess, GLU_TESS_VERTEX_DATA, (_GLUfuncptr)&vertexCell);
      gluTessCallback(tess, GLU_TESS_COMBINE_DATA, (_GLUfuncptr)&tessCombineCell);
      gluTessProperty(tess, GLU_TESS_WINDING_RULE, GLU_TESS_WINDING_POSITIVE);
    }
    else
    {
      gluTessCallback(tess, GLU_TESS_VERTEX_DATA, (_GLUfuncptr)&vertexContour);
      gluTessCallback(tess, GLU_TESS_COMBINE_DATA, (_GLUfuncptr)&tessCombineContour);
      gluTessProperty(tess, GLU_TESS_WINDING_RULE, GLU_TESS_WINDING_ODD);
    }
    // Just forces triangle-only output
    gluTessCallback(tess, GLU_TESS_EDGE_FLAG, (_GLUfuncptr)&edgeFlagVA);

    gluTessProperty(tess, GLU_TESS_BOUNDARY_ONLY, GL_FALSE);
    return tess;
  }

  void init()
  {
    GLenum gerr = glewInit();
    if(GLEW_OK != gerr)
    {
      err << "Error, cannot initialize GLEW." << endl;
      err << (char*)glewGetErrorString(gerr) << endl;
      return;
    }
    if(!GLEW_ARB_vertex_buffer_object)
    {
      err << "Error, there is no support for vertex buffer object" << endl;
      return;
    }
    glGenBuffersARB(NUMBER_VBOS, vbos);
    initialized = true;
    recompute = true;
  }

  void finalize()
  {
    glDeleteBuffersARB(NUMBER_VBOS, vbos);
  }

  void updateVBOs()
  {
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbos[VBO_POSITIONS]);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, positions.size()*sizeof(Point3d), &positions[0], GL_STATIC_DRAW_ARB);

    glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbos[VBO_NORMALS]);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, normals.size()*sizeof(Point3d), &normals[0], GL_STATIC_DRAW_ARB);

    glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);

    glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, vbos[VBO_CELL_INDICES]);
    glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, cell_indices.size()*sizeof(GLuint), &cell_indices[0], GL_STATIC_DRAW_ARB);

    glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, vbos[VBO_CONTOUR_INDICES]);
    glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, contour_indices.size()*sizeof(GLuint), &contour_indices[0], GL_STATIC_DRAW_ARB);

    glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
  }

  void colorVBOs()
  {
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbos[VBO_CELL_COLORS]);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, cell_colors.size()*sizeof(Colorf), &cell_colors[0], GL_STATIC_DRAW_ARB);

    glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbos[VBO_CONTOUR_COLORS]);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, contour_colors.size()*sizeof(Colorf), &contour_colors[0], GL_STATIC_DRAW_ARB);

    glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
  }

  void operator()(Viewer *viewer)
  {
    if(!initialized and !positions.empty()) return;
    if(recompute)
    {
      QMutexLocker locker(&mutex);
      updateVBOs();
      colorVBOs();
      recompute = false;
    }
    else if(recolor)
    {
      QMutexLocker locker(&mutex);
      colorVBOs();
      recolor = false;
    }
    glEnable(GL_POLYGON_OFFSET_FILL);
    glColor3f(1,1,1);
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_NORMAL_ARRAY);
    glEnableClientState(GL_COLOR_ARRAY);
    if(T->drawInsides) {
      glPolygonOffset(2,2);

      glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbos[VBO_POSITIONS]);
      glVertexPointer(3, GL_DOUBLE, 0, 0);

      glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbos[VBO_NORMALS]);
      glNormalPointer(GL_DOUBLE, 0, 0);

      glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbos[VBO_CELL_COLORS]);
      glColorPointer(4, GL_FLOAT, 0, 0);

      glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, vbos[VBO_CELL_INDICES]);
      glDrawElements(GL_TRIANGLES, nb_cells, GL_UNSIGNED_INT, 0);

      glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
      glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
    }

    if(T->drawBorders) {
      glPolygonOffset(-2,-2);

      glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbos[VBO_POSITIONS]);
      glVertexPointer(3, GL_DOUBLE, 0, 0);

      glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbos[VBO_NORMALS]);
      glNormalPointer(GL_DOUBLE, 0, 0);

      glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbos[VBO_CONTOUR_COLORS]);
      glColorPointer(4, GL_FLOAT, 0, 0);

      glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, vbos[VBO_CONTOUR_INDICES]);
      glDrawElements(GL_TRIANGLES, nb_contours, GL_UNSIGNED_INT, 0);

      glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
      glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
    }
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_NORMAL_ARRAY);
    glDisableClientState(GL_COLOR_ARRAY);

    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    if(pmin.x() < pmax.x())
      viewer->setSceneBoundingBox(Vec(pmin), Vec(pmax));
    else
      out << "Problem: pmin > pmax => " << pmin << " > " << pmax << endl;
  }

protected:
  void drawCellWithContour(GLUtesselator *tess_cell, GLUtesselator *tess_contour, const cell& c,
                           GLuint& cur_pos)
  {
    gluTessBeginPolygon(tess_cell, this);
    gluTessBeginPolygon(tess_contour, this);
    int n = T->S.valence(c);
    std::vector<int> pf(2*n);
    int i = 0;
    Colorf cell_color = model->cellColor(c);
    Colorf contour_color = model->contourColor(c);

    gluTessBeginContour(tess_contour);
    for(typeof(util::ForAll::range(T->S.neighbors(c))) rangeb0574d8f_b4b4_11e1_add1_bcaec51648ac = util::ForAll::range(T->S.neighbors(c)) ; !rangeb0574d8f_b4b4_11e1_add1_bcaec51648ac.atEnd() ; ++rangeb0574d8f_b4b4_11e1_add1_bcaec51648ac) {
            const junction& v = *rangeb0574d8f_b4b4_11e1_add1_bcaec51648ac.current();
#line 310 "drawing.vvh"
{
      positions[cur_pos] = model->position(v);
      cell_index[cur_pos] = c;
      normals[cur_pos] = model->normal(v);
      contour_colors[cur_pos] = contour_color;
      pf[i] = cur_pos;
      gluTessVertex(tess_contour, const_cast<GLdouble*>(positions[cur_pos].c_data()), (void*)&pf[i]);
      ++cur_pos; ++i;
    }}
    gluTessEndContour(tess_contour);

    gluTessBeginContour(tess_contour);
    gluTessBeginContour(tess_cell);
    for(typeof(util::ForAll::range(T->S.neighbors(c))) rangeb0579bae_b4b4_11e1_8589_bcaec51648ac = util::ForAll::range(T->S.neighbors(c)) ; !rangeb0579bae_b4b4_11e1_8589_bcaec51648ac.atEnd() ; ++rangeb0579bae_b4b4_11e1_8589_bcaec51648ac) {
            const junction& v = *rangeb0579bae_b4b4_11e1_8589_bcaec51648ac.current();
#line 324 "drawing.vvh"
{
      const junction& pv = T->S.prevTo(c,v);
      const junction& nv = T->S.nextTo(c,v);
      Point3d pos = findCorner(model->position(pv), model->position(v), model->position(nv), model->normal(v));
      positions[cur_pos] = pos;
      cell_index[cur_pos] = c;
      normals[cur_pos] = model->normal(c);
      cell_colors[cur_pos] = cell_color;
      contour_colors[cur_pos] = contour_color;
      pf[i] = cur_pos;
      gluTessVertex(tess_contour, const_cast<GLdouble*>(positions[cur_pos].c_data()), (void*)&pf[i]);
      gluTessVertex(tess_cell, const_cast<GLdouble*>(positions[cur_pos].c_data()), (void*)&pf[i]);
      ++cur_pos; ++i;
    }}
    gluTessEndContour(tess_contour);
    gluTessEndContour(tess_cell);

    gluTessEndPolygon(tess_contour);
    gluTessEndPolygon(tess_cell);
  }

  void drawCellNoContour(GLUtesselator *tess_cell, const cell& c,
                         GLuint& cur_pos)
  {
    gluTessBeginPolygon(tess_cell, this);
    gluTessBeginContour(tess_cell);
    int n = T->S.valence(c);
    std::vector<int> pf(n);
    int i = 0;
    Colorf cell_color = model->cellColor(c);
    for(typeof(util::ForAll::range(T->S.neighbors(c))) rangeb05810de_b4b4_11e1_9852_bcaec51648ac = util::ForAll::range(T->S.neighbors(c)) ; !rangeb05810de_b4b4_11e1_9852_bcaec51648ac.atEnd() ; ++rangeb05810de_b4b4_11e1_9852_bcaec51648ac) {
            const junction& v = *rangeb05810de_b4b4_11e1_9852_bcaec51648ac.current();
#line 355 "drawing.vvh"
{
      positions[cur_pos] = model->position(v);
      cell_index[cur_pos] = c;
      normals[cur_pos] = model->normal(v);
      cell_colors[cur_pos] = cell_color;
      pf[i] = cur_pos;
      gluTessVertex(tess_cell, const_cast<GLdouble*>(positions[cur_pos].c_data()), (void*)&pf[i]);
      ++cur_pos; ++i;
    }}
    gluTessEndContour(tess_cell);
    gluTessEndPolygon(tess_cell);
  }


  bool lineLineIntersection(Point3d& i,
                            const Point3d& p1, const Point3d& u1,
                            const Point3d& p2, const Point3d& u2,
                            const Point3d& normal)
  {
    Point3d n2 = normal ^ u2;
    double u1n2 = u1*n2;
    if(fabs(u1n2) < VVE_EPSILON) return false;
    double lambda = (p2-p1)*n2/u1n2;
    i = p1+lambda*u1;
    return true;
  }

  Point3d findCorner(const Point3d& jppos, const Point3d& jpos, const Point3d& jnpos,
                     const Point3d& normal)
  {
    Point3d up = jppos - jpos;
    up = util::normalized(up - (up*normal)*normal);
    Point3d np = normal ^ up; //(up.y(), -up.x(), 0);

    Point3d un = util::normalized(jnpos - jpos);
    un = util::normalized(un - (un*normal)*normal);
    Point3d nn = un ^ normal; //(-un.y(), un.x(), 0);

    Point3d corner;

    bool has_corner = lineLineIntersection(corner,
                                           jpos + T->cellWallWidth*nn, un,
                                           jpos + T->cellWallWidth*np, up,
                                           normal);
    if(!has_corner)
    {
      corner = jpos + T->cellWallWidth*nn;
    }

    return corner;
  }


  static void vertexCell(GLuint *v, TissueDrawer *td)
  {
    td->cell_indices.push_back(*v);
  }

  static void vertexContour(GLuint *v, TissueDrawer *td)
  {
    td->contour_indices.push_back(*v);
  }

  static void tessCombineCell(GLdouble coords[3], GLuint *vertex_data[4],
                              GLfloat w[4], GLuint **outData,
                              TissueDrawer *td)
  {
    Point3d pos = Point3d(coords);
    Point3d normal = util::normalized((w[0] > 0 ? w[0] * td->normals[*vertex_data[0]] : Point3d()) +
                                      (w[1] > 0 ? w[1] * td->normals[*vertex_data[1]] : Point3d()) +
                                      (w[2] > 0 ? w[2] * td->normals[*vertex_data[2]] : Point3d()) +
                                      (w[3] > 0 ? w[3] * td->normals[*vertex_data[3]] : Point3d()));
    Colorf color = ((w[0] > 0 ? w[0] * td->cell_colors[*vertex_data[0]] : Colorf()) +
                    (w[1] > 0 ? w[1] * td->cell_colors[*vertex_data[1]] : Colorf()) +
                    (w[2] > 0 ? w[2] * td->cell_colors[*vertex_data[2]] : Colorf()) +
                    (w[3] > 0 ? w[3] * td->cell_colors[*vertex_data[3]] : Colorf()));
    GLuint next = td->positions.size();
    td->cell_colors.resize(td->positions.size());
    td->positions.push_back(pos);
    td->cell_index.push_back(td->cell_index[*vertex_data[0]]);
    td->normals.push_back(normal);
    td->combined.push_back(next);
    td->cell_colors.push_back(color);
    *outData = &td->combined.back();
  }

  static void tessCombineContour(GLdouble coords[3], GLuint *vertex_data[4],
                                 GLfloat w[4], GLuint **outData,
                                 TissueDrawer *td)
  {
    Point3d pos = Point3d(coords);
    Point3d normal = util::normalized((w[0] > 0 ? w[0] * td->normals[*vertex_data[0]] : Point3d()) +
                                      (w[1] > 0 ? w[1] * td->normals[*vertex_data[1]] : Point3d()) +
                                      (w[2] > 0 ? w[2] * td->normals[*vertex_data[2]] : Point3d()) +
                                      (w[3] > 0 ? w[3] * td->normals[*vertex_data[3]] : Point3d()));
    Colorf color = ((w[0] > 0 ? w[0] * td->contour_colors[*vertex_data[0]] : Colorf()) +
                    (w[1] > 0 ? w[1] * td->contour_colors[*vertex_data[1]] : Colorf()) +
                    (w[2] > 0 ? w[2] * td->contour_colors[*vertex_data[2]] : Colorf()) +
                    (w[3] > 0 ? w[3] * td->contour_colors[*vertex_data[3]] : Colorf()));
    GLuint next = td->positions.size();
    td->contour_colors.resize(td->positions.size());
    td->positions.push_back(pos);
    td->cell_index.push_back(td->cell_index[*vertex_data[0]]);
    td->normals.push_back(normal);
    td->combined.push_back(next);
    td->contour_colors.push_back(color);
    *outData = &td->combined.back();
  }

  static void tessError(GLenum err)
  {
    std::cout << "Error during tesselation: ";
    switch(err)
    {
      case GLU_TESS_MISSING_BEGIN_POLYGON:
        std::cout << "Missing begin polygon";
        break;
      case GLU_TESS_MISSING_END_POLYGON:
        std::cout << "Missing end polygon";
        break;
      case GLU_TESS_MISSING_BEGIN_CONTOUR:
        std::cout << "Missing begin contour";
        break;
      case GLU_TESS_MISSING_END_CONTOUR:
        std::cout << "Missing end contour";
        break;
      case GLU_TESS_COORD_TOO_LARGE:
        std::cout << "Coordinates too large";
        break;
      case GLU_TESS_NEED_COMBINE_CALLBACK:
        std::cout << "Need combine callback";
        break;
      case GLU_OUT_OF_MEMORY:
        std::cout << "Out of memory";
        break;
      default:
        std::cout << "Unknown";
    }
    std::cout << std::endl;
  }

  static void edgeFlagVA(GLboolean)
  { }

  std::unordered_multimap<cell,GLuint> cell_positions;
  std::vector<cell> cell_index;
  std::vector<Colorf> cell_colors;
  std::vector<Colorf> contour_colors;
  std::vector<GLuint> cell_indices;
  std::vector<GLuint> contour_indices;
  std::vector<Point3d> positions;
  std::vector<Point3d> normals;
  std::list<GLuint> combined;
  Point3d pmin, pmax;
  GLuint vbos[NUMBER_VBOS];
  GLuint nb_cells, nb_contours;

  Model *model;
  const Tissue* T;
  bool recompute;
  bool recolor;
  bool withContour;
  bool initialized;
  QMutex mutex;
};

#endif // DRAWING_VVH

